"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Options(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    MAX_COUNT_FIELD_NUMBER: builtins.int
    FIXED_COUNT_FIELD_NUMBER: builtins.int
    MAX_SIZE_FIELD_NUMBER: builtins.int
    FIXED_SIZE_FIELD_NUMBER: builtins.int
    USE_CALLBACK_FIELD_NUMBER: builtins.int
    # Maximum number of entries for repeated fields.
    max_count: builtins.int = ...
    # Use a fixed length std::array for repeated fields instead of pw::Vector.
    fixed_count: builtins.bool = ...
    # Maximum size for bytes and string fields.
    max_size: builtins.int = ...
    # Use a fixed size std::array for bytes fields instead of pw::Vector.
    fixed_size: builtins.bool = ...
    # Force the use of a callback function for the field.
    use_callback: builtins.bool = ...
    def __init__(self,
        *,
        max_count : builtins.int = ...,
        fixed_count : builtins.bool = ...,
        max_size : builtins.int = ...,
        fixed_size : builtins.bool = ...,
        use_callback : builtins.bool = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"fixed_count",b"fixed_count",u"fixed_size",b"fixed_size",u"max_count",b"max_count",u"max_size",b"max_size",u"use_callback",b"use_callback"]) -> None: ...
global___Options = Options
