"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Enum(_Enum, metaclass=_EnumEnumTypeWrapper):
    pass
class _Enum:
    V = typing.NewType('V', builtins.int)
class _EnumEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Enum.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    FOO = Enum.V(0)
    BAR = Enum.V(1)

FOO = Enum.V(0)
BAR = Enum.V(1)
global___Enum = Enum

